"""
FastAPI Server Builder
"""
import json
import asyncio
from enum import Enum
from typing import Optional, List, Dict, Union, Type, Any, Sequence, Callable
from fastapi import Request, Response, APIRouter, params
from fastapi.responses import JSONResponse
from fastapi.routing import BaseRoute, APIRoute, ASGIApp, Lifespan, Default, generate_unique_id
from ._server import ServerBuilder
from ..base.meta import OBORBase


class FastAPIServerBuilder(ServerBuilder):
    """
    Dedicated RPC Server Builder for FastAPI
    """
    def create_remote_responder(
        self,
        instance: OBORBase,
        router: APIRouter,
        class_name: str,
        method_name: str,
        method: Callable
    ): # pylint: disable=too-many-arguments
        @router.post(f"{router.prefix}/{class_name}/{method_name}")
        def final_func(request: Request):
            request_body = asyncio.run(request.body())
            if request_body:
                body = json.loads(json.loads(request_body.decode()))
            else:
                body = {}
            return self.dispatch_rpc_request(instance, method, body)

    def create_remote_responder_async(
        self,
        instance: OBORBase,
        router: APIRouter,
        class_name: str,
        method_name: str,
        method: Callable
    ): # pylint: disable=too-many-arguments
        @router.post(f"{router.prefix}/{class_name}/{method_name}")
        async def final_func(request: Request):
            request_body = await request.body()
            if request_body:
                body = json.loads(json.loads(request_body.decode()))
            else:
                body = {}
            return await self.dispatch_rpc_request_async(instance, method, body)

    def build_router_from_instance(
        self,
        instance: OBORBase,
        *,
        prefix: str = "",
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        routes: Optional[List[BaseRoute]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        dependency_overrides_provider: Optional[Any] = None,
        route_class: Type[APIRoute] = APIRoute,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        lifespan: Optional[Lifespan[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(generate_unique_id),
        secure_build: bool = True,
    ): # pylint: disable=too-many-arguments,too-many-locals
        """
        build FastAPI API Router from oborpc instance
        """
        router = APIRouter(
            prefix=prefix,
            tags=tags,
            dependencies=dependencies,
            default_response_class=default_response_class,
            responses=responses,
            callbacks=callbacks,
            routes=routes,
            redirect_slashes=redirect_slashes,
            default=default,
            dependency_overrides_provider=dependency_overrides_provider,
            route_class=route_class,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            lifespan=lifespan,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            generate_unique_id_function=generate_unique_id_function
        )

        self.setup_server_rpc(instance, router, secure_build)

        return router



"""
Client RPC Builder
"""
import inspect
import json
import logging
import time
import httpx
from ..security import BASIC_AUTH_TOKEN
from ..exception import OBORPCBuildException, RPCCallException


class ClientBuilder:
    """
    Client Builder
    """
    __registered_base = set()

    def __init__(self, host, port=None, timeout=1, retry=0) -> None:
        self.master_instances = []
        self.host = host
        self.port = port
        self.timeout = timeout
        self.retry = retry

        protocol = "http://"
        if self.check_has_protocol(host):
            protocol = ""

        self.base_url = f"{protocol}{host}"
        if port:
            self.base_url += f":{port}"

        # request client
        headers = {
            "Authorization": f"Basic {BASIC_AUTH_TOKEN}",
            "Content-Type": "application/json"
        }
        self.request_client = httpx.Client(
            base_url=self.base_url,
            headers=headers
        )
        self.async_request_client = httpx.AsyncClient(
            base_url=self.base_url,
            headers=headers
        )

    def check_has_protocol(self, host: str):
        """
        Check whether the given host already defined with protocol or not
        """
        if host.startswith("http://"):
            return True
        if host.startswith("https://"):
            return True
        return False

    def check_registered_base(self, base: str):
        """
        Check whether the base RPC class is already built
        """
        if base in ClientBuilder.__registered_base:
            msg = f"Failed to build client RPC {base} : base class can only built once"
            raise OBORPCBuildException(msg)
        ClientBuilder.__registered_base.add(base)

    def create_remote_caller(
        self,
        class_name: str,
        method_name: str,
        url_prefix: str,
        timeout: float = None,
        retry: int = None
    ): # pylint: disable=too-many-arguments
        """
        create remote caller
        """
        def remote_call(*args, **kwargs):
            """
            remote call wrapper
            """
            start_time = time.time()
            try:
                data = {
                    "args": args[1:],
                    "kwargs": kwargs
                }
                url = f"{url_prefix}/{class_name}/{method_name}"
                response = self.request_client.post(
                    url=url,
                    json=json.dumps(data),
                    timeout=timeout if timeout is not None else self.timeout
                )

                if not response:
                    msg = f"rpc call failed method={method_name}"
                    raise RPCCallException(msg)

                return response.json().get("data")

            except Exception as e:
                _retry = retry if retry is not None else self.retry
                if _retry:
                    return remote_call(*args, **kwargs, retry=_retry-1)

                if isinstance(e, RPCCallException):
                    raise e
                msg = f"rpc call failed method={method_name} : {e}"
                raise RPCCallException(msg) from e

            finally:
                elapsed = f"{(time.time() - start_time) * 1000}:.2f"
                logging.debug("[RPC-Clientt] remote call take %s ms", elapsed)

        return remote_call

    def create_async_remote_caller(
        self,
        class_name: str,
        method_name: str,
        url_prefix: str,
        timeout: float = None,
        retry: int = None
    ): # pylint: disable=too-many-arguments
        """
        create async remote caller
        """
        async def async_remote_call(*args, **kwargs):
            """
            async remote call wrapper
            """
            start_time = time.time()
            try:
                data = {
                    "args": args[1:],
                    "kwargs": kwargs
                }
                url = f"{url_prefix}/{class_name}/{method_name}"
                response = await self.async_request_client.post(
                    url=url,
                    json=json.dumps(data),
                    timeout=timeout if timeout is not None else self.timeout
                )

                if not response:
                    msg = f"rpc call failed method={method_name}"
                    raise RPCCallException(msg)

                return response.json().get("data")

            except Exception as e:
                _retry = retry if retry is not None else self.retry
                if _retry:
                    return await async_remote_call(*args, **kwargs, retry=_retry-1)

                if isinstance(e, RPCCallException):
                    raise e
                msg = f"rpc call failed method={method_name} : {e}"
                raise RPCCallException(msg) from e

            finally:
                elapsed = f"{(time.time() - start_time) * 1000}:.2f"
                logging.debug("[RPC-Clientt] remote call take %s ms", elapsed)

        return async_remote_call


    def build_client_rpc(self, instance: object, url_prefix: str = ""):
        """
        Setup client rpc
        """
        _class = instance.__class__
        iterator_class = _class

        self.check_registered_base(_class)

        for (name, _) in inspect.getmembers(iterator_class, predicate=inspect.isfunction):
            if name not in iterator_class.__oborprocedures__:
                continue
            setattr(_class, name, self.create_remote_caller(_class.__name__, name, url_prefix))

    def build_async_client_rpc(self, instance: object, url_prefix: str = ""):
        """
        Setup async client rpc
        """
        _class = instance.__class__
        iterator_class = _class

        self.check_registered_base(_class)

        for (name, _) in inspect.getmembers(iterator_class, predicate=inspect.isfunction):
            if name not in iterator_class.__oborprocedures__:
                continue
            setattr(_class, name, self.create_async_remote_caller(_class.__name__, name, url_prefix))



"""
Server Builder Base
"""
import inspect


class ServerBuilder:
    """
    Server Builder
    """
    def create_remote_responder(self, instance, router, class_name, method_name, method): # pylint: disable=too-many-arguments
        """
        Remote RPC Request Responder
        """
        raise NotImplementedError("method should be overridden")

    def create_remote_responder_async(self, instance, router, class_name, method_name, method): # pylint: disable=too-many-arguments
        """
        Remote RPC Request Responder Async
        """
        raise NotImplementedError("method should be overridden")

    def dispatch_rpc_request(self, instance, method, body):
        """
        Dispatch RPC Request
        """
        args = body.get("args", [])
        kwargs = body.get("kwargs", {})
        res = method(instance, *args, **kwargs)
        return {"data": res}

    async def dispatch_rpc_request_async(self, instance, method, body):
        """
        Dispatch RPC Request
        """
        args = body.get("args", [])
        kwargs = body.get("kwargs", {})
        res = await method(instance, *args, **kwargs)
        return {"data": res}

    def setup_server_rpc(self, instance: object, router, secure_build: bool = True):
        """
        Setup RPC Server
        """
        _class = instance.__class__
        method_map = { # pylint: disable=unnecessary-comprehension
            name: method for (name, method) in inspect.getmembers(
                _class, predicate=inspect.isfunction
            )
        }

        iterator_class = instance.__class__.__base__
        iterator_method_map = { # pylint: disable=unnecessary-comprehension
            name: method for (name, method) in inspect.getmembers(
                iterator_class, predicate=inspect.isfunction
            )
        }

        for (name, method) in inspect.getmembers(iterator_class, predicate=inspect.isfunction):
            if name not in iterator_class.__oborprocedures__:
                continue

            # validate
            method = method_map.get(name)
            iterator_method = iterator_method_map.get(name)
            if secure_build:
                self.validate_implementation(name, method, _class, iterator_method, iterator_class)

            # build router
            if inspect.iscoroutinefunction(method):
                self.create_remote_responder_async(
                    instance, router, iterator_class.__name__,
                    name, method
                )
            else:
                self.create_remote_responder(
                    instance, router, iterator_class.__name__,
                    name, method
                )

    def validate_implementation(
        self,
        method_name,
        implementation_method,
        implementation_class,
        origin_method,
        origin_class,
    ):
        # validate implementation: check overridden procedure
        method_str = str(implementation_method)
        method_origin = method_str[9:method_str.find(" at 0x")].split(".")[0].strip()
        implementation_origin = str(implementation_class)[8:-2].split(".")[-1].strip()
        err = f"Unable to build. Procedure `{implementation_origin}.{method_name}()` is not implemented"
        assert method_origin == implementation_origin, err

        # validate implementation: check procedure has the same callable type
        is_implementation_coroutine = inspect.iscoroutinefunction(implementation_method)
        is_origin_coroutine = inspect.iscoroutinefunction(origin_method)
        callable_type = ["def", "async def"]
        iterator_origin = str(origin_class)[8:-2].split(".")[-1].strip()
        err = (
            f"Unable to build. Procedure `{implementation_origin}.{method_name}()` is implemented as `{callable_type[int(is_implementation_coroutine)]}`. "
            f"While the origin `{iterator_origin}.{method_name}()` is defined as `{callable_type[int(is_origin_coroutine)]}`."
        )
        assert is_implementation_coroutine == is_origin_coroutine, err
